import ast

import astcheck
import astsearch

from io import StringIO
import tokenize

from ..code import Parameter

__all__ = ['extract_definitions', 'build_definitions']

def check_list(node):
    def bool_check(node):
        return isinstance(node, ast.NameConstant) and (node.value in (True, False))
    def neg_check(node):
        return isinstance(node.operand, ast.Num) if isinstance(node, ast.UnaryOp) else False
    return all([(isinstance(n, (ast.Num, ast.Str)) 
                 or bool_check(n) or neg_check(n)) for n in node.elts])

def check_fillable_node(node, path):
    if isinstance(node, (ast.Num, ast.Str)):
        return
    elif isinstance(node, ast.UnaryOp) and isinstance(node.operand, ast.Num):
        return
    elif (isinstance(node, ast.List) 
          and isinstance(node.ctx, ast.Load) and check_list(node)):
        return
    elif isinstance(node, ast.NameConstant) and (node.value in (True, False)):
        return
    elif isinstance(node, ast.List):
        for n in node.elts:
            check_fillable_node(n, path)
        return
    elif isinstance(node, ast.Dict):
        for n in node.keys:
            check_fillable_node(n, path)
        for n in node.values:
            check_fillable_node(n, path)
        return
    
    raise astcheck.ASTMismatch(path, node, 'number, string, boolean, list or dict')

definition_pattern = ast.Assign(targets=[ast.Name()], value=check_fillable_node)

def type_and_value(node, comments={}):
    comment = comments.get(node.lineno, None)
    if isinstance(node, ast.Num):
        # int or float
        return type(node.n), node.n, comment
    elif isinstance(node, ast.Str):
        return str, node.s, comment
    elif isinstance(node, ast.List):
        return list, [type_and_value(n)[1] for n in node.elts], comment
    elif isinstance(node, ast.NameConstant) and (node.value in (True, False)):
        return bool, node.value, comment
    elif isinstance(node, ast.Dict):
        return dict, {type_and_value(node.keys[i])[1]: type_and_value(node.values[i])[1] for i in range(len(node.keys))}, comment
    elif isinstance(node, ast.UnaryOp):
        def apply_op(v, op):
            if isinstance(op, ast.USub):
               return -v
            elif isinstance(op, ast.UAdd):
               return v
            elif isinstance(op, ast.Not):
               return not v
            elif isinstance(op, ast.Invert):
               return ~v
        return type(node.operand.n), apply_op(node.operand.n, node.op), comment
    return bool, node.value, comment

def extract_comments(cell: str):
    comments = {}
    tokens = tokenize.generate_tokens(StringIO(cell).readline)
    for ttype, tstr, rowcol, _, _ in tokens:
        if ttype == tokenize.COMMENT:
           comments[rowcol[0]] = tstr
    return comments

def extract_definitions(cell):
    cell_ast = ast.parse(cell)
    comments = extract_comments(cell)
    for assign in astsearch.ASTPatternFinder(definition_pattern).scan_ast(cell_ast):
        typ, val, comment = type_and_value(assign.value, comments)
        yield Parameter(assign.targets[0].id, typ, val, comment=comment)

def build_definitions(inputs, comments=True):
    defs = []
    for param in inputs:
        s = f"{param.name} = {param.value!r}"
        if comments and param.comment:
            comment = param.comment if param.comment.startswith('#') \
                else '# ' + param.comment.lstrip()
            s +=     f"  {comment}"
        defs.append(s)
    return "\n".join(defs)



#!/usr/bin/env python3
"""Example of using nbparameterise API to substitute variables in 'batch mode'
"""

from nbclient import execute
import nbformat
from nbparameterise import extract_parameters, parameter_values, replace_definitions

stock_names = ['AAPL', 'MSFT', 'GOOG']

nb = nbformat.read("Stock display.ipynb", as_version=4)

orig_parameters = extract_parameters(nb)

for name in stock_names:
    print("Running for stock", name)

    # Update the parameters and run the notebook
    params = parameter_values(orig_parameters, stock=name)
    new_nb = replace_definitions(nb, params)
    execute(new_nb)

    # Save
    with open("Stock display %s.ipynb" % name, 'w') as f:
        nbformat.write(new_nb, f)



import copy
import importlib
import re

from nbconvert.preprocessors import ExecutePreprocessor

class Parameter(object):
    def __init__(self, name, vtype, value=None, metadata=None, comment=None):
        self.name = name
        self.type = vtype
        self.value = value
        self.metadata = metadata or {}
        self.comment = comment

    def __repr__(self):
        params = [repr(self.name), self.type.__name__]
        if self.value is not None:
            params.append(f"value={self.value!r}")
        if self.metadata:
            params.append(f"metadata={self.metadata!r}")
        if self.comment:
            params.append(f"comment={self.comment!r}")
        return "Parameter(%s)" % ", ".join(params)

    def with_value(self, value):
        """Returns a copy with value set to a new value."""
        return type(self)(
            self.name, self.type, value,  self.metadata or None, self.comment
        )

    def __eq__(self, other):
        if isinstance(other, Parameter):
            return (
                self.name == other.name
                and self.type == other.type
                and self.value == other.value
            )

def first_code_cell(nb):
    for cell in nb.cells:
        if cell.cell_type == 'code':
            return cell

kernel_name_re = re.compile(r'\w+$')

def get_driver_module(nb, override=None):
    if override:
        module_name = override
    else:
        module_name = nb.metadata.get('language_info', {}).get('name', 'python')
    assert kernel_name_re.match(module_name)
    return importlib.import_module('nbparameterise.code_drivers.%s' % module_name)

def extract_parameters(nb, lang=None):
    """Returns a list of Parameter instances derived from the notebook.

    This looks for assignments (like 'n = 50') in the first code cell of the
    notebook. The parameters may also have some metadata stored in the notebook
    metadata; this will be attached as the .metadata instance on each one.

    lang may be used to override the kernel name embedded in the notebook. For
    now, nbparameterise only handles 'python3' and 'python2'.
    """
    drv = get_driver_module(nb, override=lang)
    params = list(drv.extract_definitions(first_code_cell(nb).source))

    # Add extra info from notebook metadata
    for param in params:
        param.metadata  = nb.metadata.get('parameterise', {}).get(param.name, {})

    return params

def parameter_values(params, **kwargs):
    """Return a copy of the parameter list, substituting values from kwargs.

    Usage example::

        params = parameter_values(params,
            stock='GOOG',
            days_back=300
        )

    Any parameters not supplied will keep their original value.
    """
    res = []
    for p in params:
        if p.name in kwargs:
            res.append(p.with_value(kwargs[p.name]))
        else:
            res.append(p)
    return res

def replace_definitions(nb, values, execute=False, execute_resources=None,
                        lang=None, *, comments=True):
    """Return a copy of nb with the first code cell defining the given parameters.

    values should be a list of Parameter objects (as returned by extract_parameters),
    with their .value attribute set to the desired value.

    If execute is True, the notebook is executed with the new values.
    execute_resources is passed to nbconvert.ExecutePreprocessor; it's a dict,
    and if possible should contain a 'path' key for the working directory in
    which to run the notebook.

    lang may be used to override the kernel name embedded in the notebook. For
    now, nbparameterise only handles 'python3' and 'python2'.

    If comment is True, comments attached to the parameters will be included
    in the replaced code, on the same line as the definition.
    """
    nb = copy.deepcopy(nb)
    drv = get_driver_module(nb, override=lang)
    first_code_cell(nb).source = drv.build_definitions(values, comments=comments)
    if execute:
        resources = execute_resources or {}
        nb, resources = ExecutePreprocessor().preprocess(nb, resources)
    return nb



"""Generate an HTML form based on notebook parameters.

Used by webapp.py
"""

import htmlgen

class Input(htmlgen.Element):
    type = htmlgen.html_attribute('type', default='text')
    name = htmlgen.html_attribute('name')
    value = htmlgen.html_attribute('value')

    def __init__(self, type, name=None, value=None):
        super().__init__("input")
        self.type = type
        if name is not None:
            self.name = name
        if value is not None:
            self.value = value

class Checkbox(Input):
    checked = htmlgen.boolean_html_attribute('checked')

    def __init__(self, name=None, checked=False):
        super().__init__(type="checkbox", name=name)
        self.checked=checked

class WrapperDiv(htmlgen.Division):
    def __init__(self, *children, css_classes=None, id=None):
        super().__init__()
        self.extend(children)
        if css_classes:
            self.add_css_classes(*css_classes)
        if id is not None:
            self.id = id

py_type_to_html_input_type = {
    str: 'text',
    int: 'number',
    float: 'number',
}

def make_input_element(var):
    if var.type is bool:
        input_elm = Checkbox(var.name, var.value)
    else:
        input_elm = Input(py_type_to_html_input_type[var.type], var.name,
                          str(var.value))
        if var.type is float:
            input_elm.set_attribute('step', 'any')
    
    return input_elm

def build_form(parameters, nbname):
    doc = htmlgen.Document(title="{} (input)".format(nbname))
    doc.add_stylesheets("https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/css/bootstrap.min.css",
                        "https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/css/bootstrap-theme.min.css",
                        "/static/styles.css")
    doc.add_scripts("https://ajax.googleapis.com/ajax/libs/jquery/1.11.1/jquery.min.js",
                    "https://maxcdn.bootstrapcdn.com/bootstrap/3.2.0/js/bootstrap.min.js")

    h1 = htmlgen.Heading(1, htmlgen.Emphasis(nbname))
    h1.append(' input')
    doc.append_body(h1)

    class MyForm(htmlgen.Element):
        action = htmlgen.html_attribute('action')
        
        def __init__(self):
            super().__init__("form")
            self.action = '/submit'
            self.set_attribute('method', 'post')
        
        def generate_children(self):
            for v in parameters:
                name = v.metadata.get('display_name', v.name)
                namediv = WrapperDiv(name, css_classes=['field-name'])
                input_elm = make_input_element(v)
                yield WrapperDiv(namediv, input_elm, css_classes=['form-field'])
            
            submit = Input('submit')
            submit.id  = 'submit-button'
            submit.set_attribute('value', 'Run notebook')
            submit.add_css_classes('btn', 'btn-success')
            yield WrapperDiv(submit, id='submit-div')
    
    container = htmlgen.Division()
    container.add_css_classes('container')
    container.append(MyForm())
    doc.append_body(container)
    return doc



import os.path
import unittest

import nbformat
from nbparameterise import code, Parameter

samplenb = os.path.join(os.path.dirname(__file__), 'sample.ipynb')

class BasicTestCase(unittest.TestCase):
    def setUp(self):
        with open(samplenb) as f:
            self.nb = nbformat.read(f, as_version=4)

        self.params = code.extract_parameters(self.nb)

    def test_extract(self):
        assert self.params == [
            Parameter('a', str, "Some text"),
            Parameter('b', int, 12),
            Parameter('b2', int, -7),
            Parameter('c', float, 14.0),
            Parameter('d', bool, False),
            Parameter('e', list, [0, 1.0, True, "text", [0, 1]]),
            Parameter('f', dict, {0: 0, "item": True, "dict": {0: "text"}}),
        ]
        assert self.params[4].comment == '# comment:bool'
        assert self.params[6].comment == '# comment:dict'
        assert self.params[3].metadata == {'display_name': 'Sea'}

    def test_rebuild(self):
        from_form = [
            self.params[0].with_value("New text"),
            self.params[1].with_value(21),
            self.params[2].with_value(-3),
            self.params[3].with_value(0.25),
            self.params[4].with_value(True),
        ]
        nb = code.replace_definitions(self.nb, from_form, execute=False)

        assert "# comment:bool" in nb.cells[0].source

        ns = {}
        exec(nb.cells[0].source, ns)
        assert ns['a'] == "New text"
        assert ns['b'] == 21
        assert ns['b2'] == -3
        assert ns['c'] == 0.25
        assert ns['d'] == True

    def test_new_values(self):
        params = code.parameter_values(self.params,
            a = "New text",
            c = 12.0
        )

        assert [p.name for p in params] == ['a', 'b', 'b2', 'c', 'd', 'e', 'f']

        assert params[0].value == 'New text'
        assert params[1].value == 12
        assert params[3].value == 12.0
        assert isinstance(params[3].value, float)
        assert params[4].value == False


def test_parameter_repr():
    p = Parameter('days', int, 7, metadata={'foo': 'boo'}, comment='# Days to show')
    p2 = eval(repr(p))  # The repr should eval to an identical Parameter
    assert p2 == p
    assert p2.metadata == p.metadata
    assert p2.comment == p.comment



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['nbparameterise', 'nbparameterise.code_drivers']

package_data = \
{'': ['*']}

setup(name='nbparameterise',
      version='0.5',
      description='Re-run a notebook substituting input parameters in the first cell.',
      author=None,
      author_email='Thomas Kluyver <thomas@kluyver.me.uk>',
      url=None,
      packages=packages,
      package_data=package_data,
      python_requires='>=3.6',
     )



#!/usr/bin/env python3
"""Present an HTML form for the parameters of a notebook, and run it on submission.

To use this example, run::

    python3 webapp.py "Stock display.ipynb"

The form fields are not hardcoded here; they are built from the notebook, so it
should build a suitable form for any notebook you run this with. Try it with
Fibonacci.ipynb as well.
"""
import os.path
import sys

from nbclient import execute
from nbconvert.exporters import HTMLExporter
import nbformat
import tornado.ioloop
import tornado.web

from nbparameterise import extract_parameters, replace_definitions
from htmlform import build_form

static_path = os.path.join(os.path.dirname(__file__), 'static')


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(str(build_form(self.application.parameters,
                                  self.application.nbname)))

class SubmissionHandler(tornado.web.RequestHandler):
    def post(self):
        defined = []
        for v in self.application.parameters:
            if v.type is bool:
                inp = v.with_value(self.get_argument(v.name, default='off') == 'on')
            else:
                inp = v.with_value(v.type(self.get_argument(v.name)))
            defined.append(inp)

        nb = replace_definitions(self.application.nb, defined)
        nb = execute(nb, cwd=os.path.dirname(self.application.path))
        output, _ = HTMLExporter().from_notebook_node(nb)
        self.write(output)


class NbparameteriseApplication(tornado.web.Application):
    def __init__(self, path):
        self.path = path
        self.nb = nbformat.read(path, as_version=4)
        
        basename = os.path.basename(path)
        assert basename.endswith('.ipynb')
        self.nbname = basename[:-6]
        self.parameters = extract_parameters(self.nb)
        super().__init__([
            (r"/", MainHandler),
            (r"/submit", SubmissionHandler)
        ], static_path=static_path)


def main():
    application = NbparameteriseApplication(sys.argv[1])
    application.listen(3131)
    print("Visit http://localhost:3131/")
    tornado.ioloop.IOLoop.instance().start()

if __name__ == "__main__":
    main()
